from argparse import ArgumentParser
from importlib.metadata import distribution

args_parser = ArgumentParser(
    description="Sidera is a tool to manage scenario for system states."
)
args_parser.add_argument(
    "--version", action="version", version=distribution("sidera").version
)
args_parser.add_argument("--config", type=str, help="Configuration file to use.")
args_parser.add_argument(
    nargs="?", type=str, dest="scenario", help="Scenario file to execute."
)



from typing import Any

from pydantic import BaseModel, Field, field_validator, model_validator

from ..signal_func import DEFAULT_FUNC_NAME, func_keys


class Value(BaseModel):
    """Value model of signal."""

    name: str = Field(default="", title="Value Display Name")
    description: str = Field(default="", title="Value Description")
    function: str = Field(default="", title="Signal function name")
    params: list[Any] = Field(default=[], title="Signal function fixed parameters")
    value: float | int | None = Field(
        default=None,
        title="Static signal value",
    )

    @field_validator("function", mode="after")
    @classmethod
    def check_function(cls, value):
        assert (
            value is None or value in func_keys()
        ), f"'{value}' is not defined as function name."
        return value

    @model_validator(mode="after")
    def set_func_params(self):
        """Set function parameters from value."""
        if self.function == "":
            assert self.value is not None, "'function' or 'value' is required."
            self.function = DEFAULT_FUNC_NAME
            self.params = [self.value]
        else:
            assert self.value is None, "'function' and 'value' are exclusive."

        return self

    @model_validator(mode="after")
    def set_name(self):
        """Set value name from value or function."""
        if self.name == "":
            if self.value is not None:
                self.name = str(self.value)
            else:
                self.name = self.function
                if self.params:
                    self.name += f"({', '.join(map(str, self.params))})"
        return self



from .input import Input
from .output import Output
from .story import Story
from .scene import Scene
from .state import State
from .signal import Signal
from .value import Value
from .scenario import Scenario

__all__ = ["Input", "Output", "Story", "Scene", "State", "Signal", "Value", "Scenario"]



from pathlib import Path
from pydantic import BaseModel, Field, model_validator

from .signal import Signal
from .state import State


class Scenario(BaseModel):
    """Scenario model."""

    # Hint information
    _file: Path | None = None
    _dir: Path | None = None

    name: str = Field(default="", title="Scenario Name")
    description: str = Field(default="", title="Scenario Description")

    signals: list[Signal] | None = Field(default=None, title="Signal data list")
    states: list[State] | None = Field(default=None, title="State data list")

    @model_validator(mode="after")
    def set_name(self):
        """Set name if no name."""
        if not self.name:
            self.name = "(unnamed)"
        return self



from pydantic import BaseModel, Field


class Output(BaseModel):
    """Output model."""

    signal: str = Field(..., title="Output Signal")
    value: int | float | str = Field(..., title="Output Value")



from pydantic import BaseModel, ConfigDict, FilePath, Field

from .constants import DEFAULT_CONFIG_FILE


class ConfigModel(BaseModel):
    """Base class for the configuration models."""

    model_config = ConfigDict(validate_assignment=True, extra="ignore")

    def update(self, **new_data):
        for field, value in new_data.items():
            if hasattr(self, field):
                setattr(self, field, value)


class ArgsModel(ConfigModel):
    """Arguments model."""

    config: FilePath | None = Field(
        default=None,
        description=f"Path of the configuration file that replaces the default '{DEFAULT_CONFIG_FILE}'.",
    )
    scenario: FilePath | None = Field(
        default=None, description="Path of the scenario file to load."
    )


class CoreConfigModel(ConfigModel):
    """Config class for the application."""

    pass



from pluggy import HookspecMarker

from sidera.config import Config

plugin_hook_spec = HookspecMarker("sidera")


@plugin_hook_spec(historic=True)
def setup_config(config: Config):
    """Prepare the plugin configuration.
    Args:
        config (Config): Config instance.
    """


@plugin_hook_spec(historic=True)
def setup_logger(config: Config):
    """Prepare the logger before the initialization.
    Args:
        config (Config): Config instance.
    """


@plugin_hook_spec(historic=True)
async def initialize(config: Config):
    """Initialize the plugin.
    Args:
        config (Config): Config instance.
    """



from subprocess import run
from re import Pattern, search
from pathlib import Path
from tempfile import NamedTemporaryFile

LOGOUT_CONFIG = """
[sidera.logger.console]
level = "debug"
"""


def check_execution(args: str, pattern: Pattern[str] = r".*", return_code: int = 0):
    """Execute with the arguments and check the output and return code."""
    co = run(
        ["python", "src/sidera", *args.split()],
        capture_output=True,
        text=True,
        errors="ignore",
        encoding="utf-8",
    )
    assert co.returncode == return_code
    assert search(pattern, co.stdout)


def test_arg_version():
    """Test --version argument."""
    check_execution(args="--version", pattern=r"\d+\.\d+\.\d+\W*")


def test_arg_help():
    """Test --help argument."""
    check_execution(args="--help", pattern=r"usage: sidera")


def test_arg_config():
    """Test --config argument."""
    with NamedTemporaryFile(suffix=".toml") as fd:
        tmp_file = Path(fd.name)
        tmp_file.write_text(LOGOUT_CONFIG)
        check_execution(
            args=f"--config {tmp_file}",
            pattern=rf"Loaded config file: {tmp_file}",
        )
    check_execution(args=f"--config {tmp_file}", return_code=1)



from asyncio import sleep
from importlib.metadata import distribution
from logging import getLogger

from sidera.config import Config
from sidera.plugin import PluginHooks
from sidera.scenario import ScenarioFactory
from sidera.task import run_all_tasks, setup_task

log = getLogger(__name__)


def main():
    config = Config()
    plugin_hooks = PluginHooks(config)
    plugin_hooks.initialize()

    async def _main_entry():
        log.info("===== starting Sidera v%s =====", distribution("sidera").version)
        plugins = plugin_hooks.get_plugins()
        if config.files:
            log.info(
                "Loaded config file: %s", ", ".join(str(file) for file in config.files)
            )
        else:
            log.info("No config file loaded.")
        log.info("Plug-in[%d]: %s", len(plugins), ", ".join(plugins))

        factory = ScenarioFactory()
        if config.args.scenario:
            factory.load(config.args.scenario)
            factory.create()

        await sleep(1)

    setup_task(_main_entry(), "main")
    run_all_tasks()


if __name__ == "__main__":
    main()



import json
from pathlib import Path

from .models.scenario import Scenario

DEFAULT_SCENARIO_FILE = "scenario.json"
DEFAULT_SIGNALS_FILE = "signals.json"
DEFAULT_STATES_FILE = "states.json"


class ScenarioFactory:
    """Scenario factory class."""

    def __init__(self, src: Path | str | dict = {}):
        self.path = None
        self.data = {}
        self.root = Scenario()
        match src:
            case dict():
                self.data = src
            case Path():
                self.load(src)
            case str():
                self.data = json.loads(src)
            case _:
                raise ValueError(f"Unsupported type: {type(src)}")

    def load(self, path: Path = Path(DEFAULT_SCENARIO_FILE)):
        """Load scenario data from file or directory."""
        if not isinstance(path, Path) or not path.exists():
            raise FileNotFoundError(f"Scenario path is not found: {path}")

        if path.is_dir():
            file = path / DEFAULT_SCENARIO_FILE
            if not file.exists():
                self.load_dir(path)
                return
            path = file

        self.data = self._load_file(path, "Scenarios")
        self.path = path

        if not self.data.get("signals"):
            self.data["signals"] = self._load_file(
                path.parent / DEFAULT_SIGNALS_FILE, "Signals"
            )
        if not self.data.get("states"):
            self.data["states"] = self._load_file(
                path.parent / DEFAULT_STATES_FILE, "States"
            )

    def load_dir(self, path: Path):
        self.data["signals"] = self._load_file(path / DEFAULT_SIGNALS_FILE, "Signals")
        self.data["states"] = self._load_file(path / DEFAULT_STATES_FILE, "States")

    @staticmethod
    def _load_file(file: Path, typ: str) -> dict:
        if file.exists() and file.is_file():
            text = file.read_text(encoding="utf-8", errors="ignore")
            return json.loads(text)
        else:
            raise FileNotFoundError(f"{typ} file is not found: {file}")

    def create(self):
        """Create scenario."""
        self.root = Scenario(**self.data)
        return self.root



from argparse import ArgumentParser
from logging import StreamHandler, getLogger, Formatter
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Literal
from sys import stdout

from pydantic import Field, field_validator

from sidera.config import Config, ConfigModel
from .hook_impl import plugin_hook_impl

log = getLogger(__name__)

# Define log level type
LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"]
LOG_LEVELS = LogLevel.__args__

CONSOLE_LOG_FORMAT = "%(message)s"
FILE_LOG_FORMAT = "%(asctime)s [%(levelname)s] %(filename)s %(lineno)s: %(message)s"


@plugin_hook_impl
def setup_config(config: Config):
    """Prepare the plugin configuration."""
    add_args_parser(config.args_parser)
    config.add_config_model("logger", LoggerConfig)


@plugin_hook_impl
def setup_logger(config: Config):
    """Prepare the logger before the initialization."""
    conf: LoggerConfig = config.get_config("logger", LoggerConfig())
    root = getLogger()

    if conf.console.level:
        console_handler = StreamHandler(stream=stdout)
        console_handler.setLevel(conf.console.level)
        console_handler.setFormatter(Formatter(CONSOLE_LOG_FORMAT))
        root.addHandler(console_handler)
        root.setLevel(min(root.level, console_handler.level))

    if conf.file.level and conf.file.path:
        file_handler = RotatingFileHandler(
            filename=conf.file.path,
            maxBytes=conf.file.max_bytes,
            backupCount=conf.file.backup_count,
            encoding="utf-8",
            errors="ignore",
        )
        file_handler.setLevel(conf.file.level)
        file_handler.setFormatter(Formatter(FILE_LOG_FORMAT))
        root.addHandler(file_handler)
        root.setLevel(min(root.level, console_handler.level))
        log.debug("Log to file [%s]: %s", conf.file.level, conf.file.path)


def str_to_upper(cls, v: str):
    """Validator to convert string to uppercase."""
    return v.upper() if isinstance(v, str) else v


class FileLoggerConfig(ConfigModel):
    """Logger configuration for file output."""

    level: LogLevel | None = Field(
        default=None,
        description=f"Log level ({', '.join(LOG_LEVELS)})",
    )
    path: Path | None = Field(alias="file", default=None, description="Log file path.")
    max_bytes: int = Field(
        alias="max-bytes",
        default=1_000_000,
        description="Maximum log file size in bytes.",
    )
    backup_count: int = Field(
        alias="backup-count", default=1, description="Number of backup log files."
    )

    _str_to_upper = field_validator("level", mode="before")(str_to_upper)


class StreamLoggerConfig(ConfigModel):
    """Logger configuration for stream output."""

    level: LogLevel | None = Field(
        default="INFO",
        description=f"Log level ({', '.join(LOG_LEVELS)})",
    )

    _str_to_upper = field_validator("level", mode="before")(str_to_upper)


class LoggerConfig(ConfigModel):
    """Logger configuration."""

    file: FileLoggerConfig = Field(
        default_factory=FileLoggerConfig, description="File logger configuration."
    )
    console: StreamLoggerConfig = Field(
        default_factory=StreamLoggerConfig, description="Stream logger configuration."
    )


def add_args_parser(parser: ArgumentParser):
    """Customize argument parser."""
    # TODO: this argument is for debug.
    parser.add_argument(
        "--logger-version", action="version", version="Simple Logger: v0.1.0"
    )



from pydantic import BaseModel, Field

from .story import Story
from .input import Input
from .output import Output


class State(BaseModel):
    """State model."""

    name: str = Field(..., title="State Name")
    description: str = Field(default="", title="State Description")
    input: list[Input] = Field(default=[], title="Input Signals")
    output: list[Output] = Field(default=[], title="Output Signals")
    stories: list[Story] = Field(default=[], title="Transition Stories")



from typing import Any, Callable
from logging import getLogger

log = getLogger(__name__)

ValueFunc = Callable[[float, int, list[Any]], float]

DEFAULT_FUNC_NAME = "static"


def static_value(prev: float, time: int, params: list[int | float]):
    """Return static value."""
    assert (
        len(params) > 0
    ), "needs params that contains at least one item, but args is empty."
    return float(params[0])


_value_func_set: dict[str, ValueFunc] = {DEFAULT_FUNC_NAME: static_value}


def add_func(name: str, func: ValueFunc):
    """Add signal function."""
    _value_func_set[name] = func
    log.debug("Add signal function: %s", name)


def get_func(name: str):
    """Get value function."""
    return _value_func_set.get(name)


def func_keys(name: str):
    """Get value function."""
    return _value_func_set.keys()



from typing import Literal
from pydantic import BaseModel, Field

WaitTime = Literal["any"] | float


class Scene(BaseModel):
    """Scene model."""

    name: str = Field(default="", title="Scene Name")
    description: str = Field(default="", title="Scene Description")
    wait_time: WaitTime = Field(default="any", title="Scene wait time")
    signal: str = Field(..., title="Changed signal name")
    value: int | float | str = Field(..., title="Changed signal value")



from pluggy import PluginManager
from sys import exit, stderr
from pydantic import ValidationError

from sidera.config import Config

from . import hook_specs
from .plugin_registry import PLUGIN_REGISTRY_LIST


class PluginHooks:
    """Plugin hooks provider."""

    def __init__(self, config: Config):
        self.config = config
        self.plugin_manager = self.get_plugin_manager()

    @staticmethod
    def get_plugin_manager() -> PluginManager:
        """Get plugin manager."""
        pm = PluginManager("sidera")
        pm.add_hookspecs(hook_specs)
        pm.load_setuptools_entrypoints("sidera")
        # resister default plugins
        for plugin in PLUGIN_REGISTRY_LIST:
            pm.register(plugin)
        return pm

    def initialize(self):
        """Initialize the plugins."""
        try:
            self.setup_config()
        except ValidationError as e:
            print(f"error: parsing configurations: {e}", file=stderr)
            exit(1)
        self.init_logger()
        self.init_plugin()

    def setup_config(self):
        """Setup the configuration customized by plugin."""
        self.plugin_manager.hook.setup_config.call_historic(
            kwargs=dict(
                config=self.config,
            )
        )
        try:
            self.config.parse_args()
        except ValidationError as e:
            self.config.print_usage_args()
            print(f"error: parsing arguments: {e}", file=stderr)
            exit(1)
        try:
            self.config.load()
        except ValidationError as e:
            print(f"error: parsing configurations: {e}", file=stderr)
            exit(1)

    def init_logger(self):
        """Setup the logger before the initialization."""
        self.plugin_manager.hook.setup_logger.call_historic(
            kwargs=dict(
                config=self.config,
            )
        )

    def init_plugin(self):
        """Initialize the plugins."""
        self.plugin_manager.hook.initialize.call_historic(
            kwargs=dict(
                config=self.config,
            )
        )

    def get_plugins(self) -> list[str]:
        """Get the list of plugin names."""
        return [name for name, _ in self.plugin_manager.list_name_plugin()]



from __future__ import annotations
from collections.abc import Coroutine
from asyncio import TaskGroup, Task
from logging import getLogger
import asyncio

log = getLogger(__name__)

_coro_list: dict[str, Worker] = {}


def setup_task(coro: Coroutine, name: str):
    """Setup the task."""
    task = Worker(coro, name)
    _coro_list[name] = task
    log.debug("Task [%s] is registered.", name)


def run_all_tasks():
    """Run all tasks."""
    log.info("All workers start.")
    try:
        asyncio.run(Worker.start())

        log.info("All workers finished.")
    except KeyboardInterrupt:
        log.warning("KeyboardInterrupt is received.")


class Worker:
    _task_list: list[Task] = []

    def __init__(self, coro: Coroutine, name: str):
        self.name = name
        self.coro = coro
        self.task: Task | None = None
        if self._task_list:
            raise RuntimeError("Task group is already running.")

    @classmethod
    async def start(cls):
        """Start the task group."""
        async with TaskGroup() as tg:
            for name, worker in _coro_list.items():
                worker.task = tg.create_task(worker.coro, name=name)
                cls._task_list.append(worker.task)
                log.debug("Task [%s] starts.", name)

        cls._task_list.clear()



from pydantic import BaseModel, Field


class Input(BaseModel):
    """Input model."""

    signal: str = Field(..., title="Input Signal")
    value: int | float | str = Field(..., title="Input Value")



from os import getcwd
from pathlib import Path
from typing import Type
from argparse import ArgumentParser
from tomllib import TOMLDecodeError, loads

from .args_parser import args_parser
from .constants import DEFAULT_CONFIG_FILE
from .models import ArgsModel, ConfigModel, CoreConfigModel, Field


class Config:
    """Configurations manager for the application and all plugins."""

    def __init__(self):
        self._files: list[Path] = []
        self._parser: ArgumentParser = args_parser
        self._config_models: dict[str, Type[ConfigModel]] = {"": CoreConfigModel}
        self._args_models: dict[str, Type[ConfigModel]] = {"core": ArgsModel}
        self._configs: dict[str, ConfigModel] = {}
        self._args: dict[str, ConfigModel] = {}

    def parse_args(self):
        """Parse the arguments"""
        args = self._parser.parse_args()
        for key, model in self._args_models.items():
            self._args[key] = model(**vars(args))

        # Add specified configuration file with args.
        if self.args.config:
            self.add_file(self.args.config)
        else:
            self.add_file(Path(getcwd()) / DEFAULT_CONFIG_FILE)

    def print_usage_args(self):
        """Print the usage of the arguments."""
        self._parser.print_usage()

    def add_file(self, file: Path) -> bool:
        """Add a configuration file to the list.
        Returns:
            bool: True if the file was added, False otherwise.
        """
        if file.exists() and file.is_file():
            self._files.append(file.expanduser())

    def load(self):
        """Load the configuration from the file."""
        for file in self._files:
            if not file.exists() or not file.is_file():
                continue

            text = file.read_text(encoding="utf-8", errors="ignore")
            try:
                conf_dict = loads(text)
                sidera_dict: dict = conf_dict.get("sidera", {})
            except TOMLDecodeError as e:
                raise TOMLDecodeError(f"Error decoding {file}: {e}") from e

            for key, model in self._config_models.items():
                if key == "":
                    update_items = sidera_dict
                else:
                    update_items = sidera_dict.get(key, {})
                origin = self._configs.get(key)
                if origin is None:
                    self._configs[key] = model(**update_items)
                else:
                    self._configs[key] = origin.update(**update_items)

    def save(self, file: Path = [Path(getcwd()) / DEFAULT_CONFIG_FILE]):
        """Save the configuration to the file."""
        # TODO: Implement the save method.
        pass

    def add_config_model(self, key: str, model: Type[ConfigModel]):
        """Add a configuration model to the manager."""
        self._config_models[key] = model

    def get_config(
        self, key: str, default: ConfigModel | None = None
    ) -> ConfigModel | None:
        """Get the configuration for each model."""
        return self._configs.get(key, default)

    @property
    def core(self) -> CoreConfigModel:
        """Return the core configuration."""
        return self._configs.get("core", CoreConfigModel())

    @property
    def args(self) -> ArgsModel:
        """Return the core arguments."""
        return self._args.get("core", ArgsModel())

    @property
    def args_parser(self) -> ArgumentParser:
        """Return the argument parser."""
        return self._parser

    @property
    def files(self) -> list[Path]:
        """Return the list of configuration files."""
        return self._files


__all__ = ["Config", "ConfigModel", "Field"]



from .main import main
from .config import Config, ConfigModel, ArgsModel, CoreConfigModel, Field
from .task import Task, setup_task

__all__ = [
    "main",
    "Config",
    "Task",
    "setup_task",
    "ConfigModel",
    "ArgsModel",
    "CoreConfigModel",
    "Field",
]



from pydantic import BaseModel, Field

from .scene import Scene


class Story(BaseModel):
    """Story model."""

    name: str = Field(default="", title="Story Name")
    description: str = Field(default="", title="Story Description")
    scenes: list[Scene] = Field(default=[], title="Scenes List")
    destination: str | None = Field(default=None, title="Destination State")



from pydantic import BaseModel, Field

from .value import Value


class Signal(BaseModel):
    name: str = Field(..., description="Signal name", min_length=1)
    description: str = Field(default="", description="Signal description")
    values: list[Value] = Field(..., description="Signal values list", min_length=1)
